/* ADH_list.h  (C) 2005  adolfo@di-mare.com  */

/** \file  ADH_list.h
    \brief Lista simple de nodos con puntero final nulo

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2005
*/


#ifndef ADH_list_h
#define ADH_list_h   ///< Evita la inclusin mltiple

//#define  NO_namespace    // omite el uso de "namespace" en ADH_port.h
#include <iostream>       // #include <iostream>
// #include "ADH_port.h"
#include "Tdef.h" // truco para usar pseudo-plantillas

class test_ADH_list;

namespace ADH { // namespace ADH {

class biderectional_iterator;

/** Lista simple de nodos con puntero final nulo.
    - Esta clase tiene varias de las operaciones ms importantes
      de la clase \c "list" de la biblioteca estndar de C++.

    - Es la que usan los estudiantes de Adolfo para aprender.
*/
class list {
    private: class node;
    public:  class iterator;
    friend   class iterator;
    friend   class biderectional_iterator;

    typedef T  value_type;  ///< Nombre estndar del objeto contenido
    typedef value_type& reference; ///< Referencia al objeto contenido
    typedef const value_type& const_reference; ///< Referencia constante al objeto contenido
    typedef unsigned size_type;  ///< Nombre estndar del tipo retornado por \c size()
private:
    /// Nodos de la clase \c "list"
    class node {
        friend class ADH::list;
        friend class ADH::list::iterator;
        friend class ADH::biderectional_iterator;

    private:
        node(const value_type& v) : m_val(v) { /* next = 0; */ } ///< Constructor para \c "v"

        friend std::ostream& operator<< (std::ostream &, const list& );
        friend bool check_ok( const list & L );
        friend bool check_ok( const biderectional_iterator & I );
    private:
        value_type  m_val; ///< Valor almacenado en el nodo
        node      * next;  ///< Puntero al siguiente nodo
    };  // node

public:
    /// Iteradores simples hacia adelante (<em>forward</em>) para la clase \c "list"
    class iterator {
        friend class ADH::list;
    public:
        iterator() : m_p(0) { /* inicializo, como los de STL */ } ///< Constructor

        /// Constructor de copia
        iterator( const iterator& o ) : m_p(o.m_p) { }  // copia el Rep directo
        iterator& operator=( const iterator& ); ///< Copia

        friend bool operator == ( const iterator&, const iterator& );
        friend bool operator != ( const iterator&, const iterator& );

        iterator operator++()
            { m_p = m_p -> next; return *this; }                   ///< \c ++p
        iterator operator++(int)
            { iterator q = (*this); m_p = m_p -> next; return q; } ///< \c p++

        value_type& operator*()  { return   m_p -> m_val ; } ///< \c *p
        #if !defined(__BORLANDC__) || (__BORLANDC__ > 0x0410)
        value_type* operator->() { return &(m_p -> m_val); } ///< \c p->
        #endif
    private:
        friend bool check_ok( const list::iterator & I );
    private:
        node * m_p;   ///< Puntero al nodo de \c "list"
    }; // list::iterator

public:
    list() : m_first(0) { } ///< Constructor de vector
    list(const list& L) : m_first(0) { *this = L; } ///< Constructor de copia
    ~list() { clear(); } ///< Destructor
    void clear(); ///< Elimina los valores del contenedor y lo deja vaco

    void push_front( const list::value_type& v );
    void push_back(  const list::value_type& v );

    void pop_front();
    void pop_back();

    bool      empty() const; ///< Retorna "true" si la lista est vaca
    size_type size()  const; ///< Cantidad de valores almacenados en \c "*this"

    const_reference front() const;
    const_reference back()  const;

    /// Denota al primer valor de la lista
    iterator begin() const { iterator p; p.m_p = m_first; return p; }
    /// Denota el valor que ya est fuera del contenedor
    iterator end  () const { iterator p; p.m_p = 0;       return p; }

    list& operator=(const list& LO); ///< Copia ==> <code>L = LO</code>
    void move(list& L);
    void swap(list& L);   // Intercambia <code>*this <==> L</code>

    void insert(iterator position, const value_type& v);
    void remove(const value_type& v);
    void splice( iterator p, list& L, iterator from );

    friend bool operator == (const list&, const list&);

    friend std::ostream& operator << (std::ostream &, const list& );
    friend std::istream& operator >> (std::istream &,       list& );

    friend bool check_ok( const list& );
    friend bool check_ok( const list::iterator & I );
    friend bool check_ok( const biderectional_iterator & I );
    friend class ::test_ADH_list; ///< Datos de prueba para la clase.

private:
    node * prev_node( const node * p ) const;
    node * last_node( ) const;

private:
    node * m_first;  ///< Puntero al primer nodo de la lista
}; // list

inline bool list::empty() const {
    return (m_first == 0);
}

/// Referencia al primer valor de la lista. \pre <code> ! empty() </code>
inline list::const_reference list::front() const {
    return m_first->m_val;
} // list::front()

/// Referencia al ltimo valor de la lista. \pre <code> ! empty() </code>
inline list::const_reference list::back() const {
    list::node * p;
    p = ( m_first == 0 ? 0 : last_node( ) );
    return p->m_val;
} // list::back()

list::iterator last( const list & L );

/// Copia
inline list::iterator& list::iterator::operator= (
    const list::iterator& o
) {
    m_p = o.m_p; // copia el Rep directo
    return *this;
}

///  <code>p == q</code> ?
inline bool operator == (
    const list::iterator& p,
    const list::iterator& q
) {
    return ( p.m_p == q.m_p );
}
///  <code>p != q</code> ?
inline bool operator != (
    const list::iterator& p,
    const list::iterator& q
) {
    return !( p == q ); // return operator == (p, q);
}

/// Verifica la invariante para la clase
inline bool check_ok( const list::iterator & I ) {
    if (I.m_p == 0) {
        return true; // "I.m_p" es un puntero nulo
    }
    list TempL; TempL.m_first = I.m_p; // simula una lista a partir de "I.m_p"
    bool ok = check_ok( TempL );
    TempL.m_first = 0; // evita que el destructor de "TempL" destruya el resto
    if (! ok) {
        return false; // lista mala ==> iterador malo
    }
    return true;
}

#include <assert.h>

/// Iteradores bidireccionales adelante && atrs para la clase \c "list".
/// - Aunque este iterador tiene el mtodo "--" sobrecargado, la
///   complejidad de ejecucin es <code>O( size() )</code>.
class biderectional_iterator {
    friend class list;
public:
    biderectional_iterator() : m_p(0), m_L(0) { /* inicializo, como los de STL */ } ///< Constructor

    /// Constructor de copia
    biderectional_iterator( const biderectional_iterator& o)
         { m_p = o.m_p; m_L = o.m_L; }  // copia el Rep directo
    biderectional_iterator& operator=( const biderectional_iterator& o)
         { m_p = o.m_p; m_L = o.m_L; return *this; } ///< Copia

    ///  <code>p == q</code> ?
    friend bool operator == ( const biderectional_iterator& p, const biderectional_iterator& q) {
        return ( p.m_p == q.m_p && p.m_L == q.m_L) ; }
    friend bool operator != ( const biderectional_iterator& p, const biderectional_iterator& q) {
        return !(p == q); } ///<  <code>p != q</code> ?

    biderectional_iterator operator++()
        { m_p = m_p -> next; return *this; }           ///< \c ++p
    biderectional_iterator operator++(int)
        { biderectional_iterator q = (*this);
          m_p = m_p -> next; return q; }               ///< \c p++

    biderectional_iterator operator--()
        { m_p = m_L->prev_node(m_p); return *this; }   ///< \c --p
    biderectional_iterator operator--(int)
        { biderectional_iterator q = (*this);
          m_p = m_L->prev_node(m_p); return q; }       ///< \c p--

    list::value_type& operator*()  { return   m_p -> m_val ; } ///< \c *p
//  list::value_type* operator->() { return &(m_p -> m_val); } ///< \c p->

private:
    list::node * m_p;   ///< Puntero al nodo de \c "list"
    list       * m_L;   ///< Puntero a la \c "list"

private:
    /** Verifica la invariante de la clase
        - Regresa \c "true" si \c "L" contiene un valor correcto.
        - Podra retornar \c "true" an cuando \c "L" tenga su valor corrupto.
        - Podra no retornar si \c "L" tiene su valor corrupto.

    \par Rep Diagrama de la clase
    \code
                                                  NULL
                +-----+---+   +-----+---+   +-----+---+
    L.m_first-->| v_1 | *-|-->| v_2 | *-|-->| v_3 | 0 |
        ^       +-----+---+   +-----+---+   +-----+---+
        |       m_val  next        ^
        |                          |
        |                          |
        |            +---+---+     |
        +-----<------|-* | *-|-----+
                     +---+---+
                      m_L m_p
                                            last_node(L)
    \endcode
    */
    friend bool check_ok( const biderectional_iterator & I ) {
        if (! check_ok( *I.m_L )) {
            return false; // lista mala ==> iterador malo
        }
        assert( I.m_L != 0 );
        if (I.m_p == 0) {
            return true; // "I.m_p" es un puntero nulo
        }
        assert( I.m_p != 0 );
        list::node * p;
        for (p = I.m_L->m_first; p != 0; ++p) {
            if (p == I.m_p) {
                break; // busca a "I.m_p" dentro de la lista "*(I.m_L)"
            }
        }
        return (p == I.m_p);
    }
}; // biderectional_iterator

/// x != y?
inline bool operator != (const list&x, const list&y) {
    return ! (x==y);
}

} // namespace ADH // } // namespace ADH

#endif // ADH_list_h

// EOF: ADH_list.h
