/* ADH_map.h  (C) 2004  adolfo@di-mare.com  */

/** \file  ADH_map.h
    \brief Versin muy simplificada del diccionario STL

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2005
*/

#ifndef ADH_map_h
#define ADH_map_h   "Versin 2"

#include <iostream>     
// #include "ADH_port.h"

#include "Tdef.h" // truco para usar pseudo-plantillas
#include "ADH_list.h"

namespace ADH { // namespace ADH {

/**  La clase diccionario permite usar vectores asociativos.
     - La operacin principal es D["index"] = 32.
*/
class map {

public:
    typedef T  value_type;  ///< Nombre estndar del objeto contenido
    typedef T1 key_type;    ///< Llave del par
    typedef T2 mapped_type; ///< Valor asociado a la llave
    typedef value_type& reference; ///< Referencia al objeto contenido
    typedef const value_type& const_reference; ///< Referencia constante al objeto contenido

    /// Mecanismo utilizado para recorrer los valores almacendos en el diccionario
    class iterator {    friend class map;
    public:
        iterator() : m_q() { /* inicializo, como los de STL */ } ///< Constructor de vector

        iterator( const iterator& o ) ///< Constructor de copia
             { m_q = o.m_q;}  // copia el Rep directo
        iterator& operator=( const iterator& ); ///< Copia <code>p = q</code>

        friend bool operator == ( const iterator&, const iterator& );
        friend bool operator != ( const iterator&, const iterator& );

        iterator operator++()                  ///< ++p
            { m_q++; return *this; }
        iterator operator++(int)               ///< p++
            { iterator qT = (*this); m_q++; return qT; }

        value_type& operator*()               ///< *p
            { return  (*m_q); }
        value_type* operator->()              ///< p->
            { return &(*m_q); }
    private:
        list::iterator m_q;   ///< Iterador sobre la lista que contiene las parejas del diccionario
    }; // map::iterator

public:
    map() : m_L() { } ///< Constructor de vector
    map(const map& MO) : m_L() { *this = MO; } ///< Constructor de copia
    ~map() { } ///< Destructor

    bool empty() const { return m_L.empty(); } ///< Retorna \c "true" si el contenedor est vaco

    map& operator=(const map& MO) {
        m_L = MO.m_L; return *this; } ///< Copia <code>DD = DD0</code>
    void move(map& M) { m_L.move(M.m_L); } ///< Traslada el valor de \c "M" a \c "*this"
    void swap(map& M) { m_L.swap(M.m_L); } ///< Intercambia los valores de \c "M" <==> \c "*this"

    iterator begin() const { iterator p; p.m_q = m_L.begin(); return p; } ///< Primer valor del contenedor
    iterator end  () const { iterator p; p.m_q = m_L.end();   return p; } ///< Lo que est "fuera" del contenedor

    iterator find(const key_type & k);  // Encuentra DDD[k] si existe
    iterator insert (const value_type &);  // Agrega un nuevo par
    void Ordene() { /* NO lo he implementado */ } ///< Es mejor usar \c "sort<>()"

    friend std::ostream& operator << ( std::ostream &, const map& );
    friend std::istream& operator >> ( std::istream &,       map& );
    friend bool check_ok( const map & DDD );

private:
    friend void printMap(const map& );

public:
    mapped_type & operator[] (const key_type &); // D["index"] = 32

private:
    list m_L; ///< Lista que contiene los valores del diccionario
}; // map

inline map::iterator& map::iterator::operator= (
    const map::iterator& o
) {
//  copia
    m_q = o.m_q; // copia el Rep directo
    return *this;
}

///  <code>p == q</code> ?
inline bool operator == (
    const map::iterator& x,
    const map::iterator& y
) {
    return x.m_q==y.m_q;
}
///  <code>p != q</code> ?
inline bool operator != (
    const map::iterator& x,
    const map::iterator& y
) {
    return !(x==y); // return operator == (x, y);
}

} // namespace ADH // } // namespace ADH

#endif // ADH_map_h

// EOF: ADH_map.h
