// Astring.h      (c) 2005 adolfo@di-mare.com

/** \file  Astring.h
    \brief Hileras con una parte de la funcionalidad de las hileras en \c "string"

    - La implementacin es muy simple y funciona, pero no es muy eficiente
      (es adecuada para los estudiantes).
    - No incluye un constructor a partir de \c "char".
    - No incluye un convertidor a \c "(const char *)".
    - Difiere de la clase \c "string" en que es imposible que la letra
      (char)0 sea alguna de las letras dentro de una hilera \c "Astring".

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2005
*/

#ifndef Astring_h
#define Astring_h // evita la inclusin mltiple

#include <iostream>     
// #include "ADH_port.h"

#include <cstdlib>
#include <cstring>
#include <climits> // UINT_MAX

/// Hileras simples implementadas como un (char*) a su valor.
/// - Todas las hileras terminan en el caracter 0 binario (con <em>C-strings</em>).
/// - La implementacin es muy simple y funciona, pero no es muy eficiente
///   (es adecuada para los estudiantes).
class Astring {
private:
    char * m_s; ///< Puntero a la hilera
    static char null;  ///< Campo esttico para representar "" ==> hilera nula
public:
    typedef char value_type;  ///< Tipo base de la hilera
    typedef value_type& reference;  ///< Tipo base de la hilera (referencia modificable)
    typedef const value_type& const_reference; ///< Tipo base de la hilera (referencia)
    Astring()                   : m_s(0) { }  ///< Constructor de vector
    /// Constructor a partir de otra hilera
    Astring(const Astring& str) : m_s(0) {  this->operator = (str.m_s); }
    /// Constructor a partir de una hilera C [terminada en \c "(char)0"]
    Astring(const char* str)    : m_s(0) { *this = str;    }

//  No se puede implementar este constructor porque produce ambigedad
//  Astring(char c) : m_s(0) { char v[2]; v[0] = c; v[1] = 0; *this = v; }

    ~Astring() { if (m_s != 0) { delete [] m_s; } }  ///< Destructor

    /// S = (char)'c';
    Astring& operator  = (const char c) { char v[2]; v[0] = c; v[1] = 0; return (*this = v); }
    Astring& operator  = (const char *); ///< S = "C string";
    Astring& operator += (const char *); ///< S += "C string";
    Astring& operator += (char); ///< S += 'c';

    Astring& operator  = (const Astring& str);
    /// Agrega \c "str" al final de \c "*this"
    Astring& operator += (const Astring& str) { return *this += str.m_s; }
    Astring& assign(const Astring& str, unsigned pos = 0, unsigned n = UINT_MAX);

    void Set(const char *str, size_t n);

    const_reference operator[] (size_t i) const { return m_s[i]; } ///< Str[i]
    reference       operator[] (size_t i)       { return m_s[i]; } ///< Str[i]

    friend Astring operator + (const Astring&, const Astring&);

    friend Astring operator + (const char,     const Astring& );
    friend Astring operator + (const Astring&, const char     );
    friend Astring operator + (const char*,    const Astring& );
    friend Astring operator + (const Astring&, const char*    );

//  No se puede implementar este convertidor porque produce ambigedad
//  operator const char* () const { return (m_s==0 ? (&null) : m_s); }

    const char* c_str() const { return (m_s==0 ? (&null) : m_s); }
    const char* data()  const { return c_str(); }

    friend bool operator == (const Astring&, const Astring&);
    friend bool operator <  (const Astring&, const Astring&);
    friend bool operator != (const Astring&, const Astring&);
    friend bool operator <= (const Astring&, const Astring&);
    friend bool operator >= (const Astring&, const Astring&);
    friend bool operator >  (const Astring&, const Astring&);

    void resize( size_t n, char ch = char() );
    unsigned size() const { return ( m_s==0 ? 0 : strlen(m_s) ); }

    friend double real   (const Astring& );   // Conversin a real
    friend long   integer(const Astring& );   // Conversin a long

    friend bool check_ok( const Astring& );
//  excluidos porque producen ambigedad con operadores aritmticos
//  operator double () { return double(m_s) / double(_den); }
//  operator long   () { return        m_s  /        _den ; }
}; // Astring


///  x == y ?
inline bool operator == (const Astring &x, const char * y) {
    return strcmp( x.c_str(), y) == 0;
}  // operator ==

///  x < y ?
inline bool operator  < (const Astring &x, const char * y) {
    return strcmp(x.c_str(), y) < 0;
}  // operator <

///  x == y ?
inline bool operator == (const Astring &x, const Astring &y) {
    return strcmp(x.c_str(), y.c_str()) == 0;
}  // operator ==

///  x < y ?
inline bool operator < (const Astring &x, const Astring &y) {
    return strcmp(x.c_str(), y.c_str()) < 0;
}  // operator <

///  x > y ?
inline bool operator > (const Astring &x, const Astring &y) {
    return (y < x);
}  // operator >

///  x != y ?
inline bool operator != (const Astring& x, const Astring& y) {
    return !(x == y);
}  // operator !=

///  x <= y ?
inline bool operator <= (const Astring& x, const Astring& y) {
    return !(y < x);
}  // operator <=

///  x >= y ?
inline bool operator >= (const Astring& x, const Astring& y) {
    return !(x < y);
}  // operator >=

inline Astring& Astring::operator += (char c) {
    char v[2]; v[0] = c; v[1] = 0;
    return *this += v;
}

/// Calcula y retorna la concatenacin \c "l + r"
inline Astring operator + (const Astring& l, const Astring& r) {
    Astring res = l;
    res += r;
    return res;
}

/// Calcula y retorna la concatenacin \c "l + r"
inline Astring operator + (const char l, const Astring& r) {
    Astring str;
    str = l;
    return str + r;
}

/// Calcula y retorna la concatenacin \c "l + r"
inline Astring operator + (const Astring& l, const char r) {
    Astring str;
    str = r;
    return l + str;
}

/// Calcula y retorna la concatenacin \c "l + r"
inline Astring operator + (const char* l, const Astring& r) {
    Astring str = l;
    return str + r;
}

/// Calcula y retorna la concatenacin \c "l + r"
inline Astring operator + (const Astring& l, const char* r) {
    Astring str = r;
    return l + str;
}

inline std::ostream& operator << ( std::ostream& COUT, const Astring& s ) {
    return COUT << s.c_str();
};


std::istream& operator >> (std::istream&,       Astring& );

/// <code> str = other_str; </code>
/// - Operador de copia
inline Astring& Astring::operator = (const Astring& str) {
    if (this != &str) { // evite borrado en auto-copia
        *this  = str.m_s;
    }
    return *this;
}

#endif // Astring_h

// EOF: Astring.h
